#!/usr/bin/env python3
"""Module: sc_tcr_utility.py"""

import logging
import sys
import subprocess
import re
import glob
import os
import pandas as pd
import numpy as np
from collections import OrderedDict

# import generic_functions
from lib.shared_utility import *

###############################################################################
# Function definitions
###############################################################################

def load_printed_barcode(infile, logname):
    """ Function: load printed barcode list in 
    comma-separated values (CSV) format
    The barcode list has a single column with barcode sequence
    """
    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info",
              "Function load_printed_barcode started", module_logname)

    printed_barcode_list = []

    try:
        with open(infile, "r") as printed_barcode_file:
            lines = printed_barcode_file.readlines()
            for line in lines:
                printed_barcode_list.append(line.rstrip().split(" ")[0])
    except EnvironmentError as err:
        write_log("error", "Unable to open file: " + infile +
                  "\n" + str(err), module_logname)

    write_log("info",
              "Function load_printed_barcode: successfully completed",
              module_logname)
    return printed_barcode_list


def load_selected_barcode(infile, logname):
    """ Function: load selected barcode from CellSelect Well list
    RowID, ColumnID and sample type would be concatenated as sample IDs, 
    and a csv format barcode list would be generated."""
    
    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info",
              "Function load_selected_barcode started", module_logname)

    selected_barcode_dict = OrderedDict()

    try:
        with open(infile, "r") as selected_barcode_file:
            next(selected_barcode_file)
            for line in selected_barcode_file:
                line_content = line.rstrip().split("\t")
                line_content[4] = line_content[4].replace(" ", "")
                line_content[4] = line_content[4].replace("-", "")
                line_content[4] = line_content[4].replace("_", "")
                selected_barcode_dict[line_content[4] + "_R" + 
                    line_content[0]+ "_C" + line_content[1]] = \
                    line_content[5]
    except EnvironmentError as err:
        write_log("error", "Unable to open file: " + infile +
                  "\n" + str(err), module_logname)

    prefix = os.path.splitext(infile)[0]
    outcsv = prefix + ".csv"
    with open(outcsv, 'w') as f:
        [f.write('{0},{1}\n'.format(key, value)) for key, value in 
         selected_barcode_dict.items()] 
    
    write_log("info",
              "Function load_selected_barcode: successfully completed",
              module_logname)
    return selected_barcode_dict


def do_barcode_classification(printed_barcode_list, selected_barcode_dict,
	                            selected_barcode_list, logname):
    """Function: take printed barcode list and selected barcode list as input
    classify notselected barcode and invalid barcode, and write barcodes of 
    three categories (notselected,selected,printed) to csv files
    Returns True if successfull."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function do_barcode_classification started ",
              module_logname)

    notselected_barcode_list = [i for i in printed_barcode_list if
                                i not in 
                                list(selected_barcode_dict.values())]

    num_notselected = len(notselected_barcode_list)

    notselected_barcode_dict = OrderedDict(
        zip([a + b for a, b in
             zip(["notselected"] * num_notselected,
                 list(map(str, range(1, num_notselected + 1))))],
            notselected_barcode_list))

    # concatenate selected & notselected dictionaries
    # full_barcode_dict = {}
    full_barcode_dict = OrderedDict()
    full_barcode_dict.update(selected_barcode_dict)
    full_barcode_dict.update(notselected_barcode_dict)
    
    with open(os.path.splitext(selected_barcode_list)[0] + 
        "_printed.csv","w") as printed:
        writer3 = csv.writer(printed)
        for row in full_barcode_dict.items():
            writer3.writerow(row)

    write_log("info",
              "Function do_barcode_classification finished", 
              module_logname)
    return full_barcode_dict


def split_fastq(demux_tool_loc, mate_fastq, barcode_fastq,
                barcode_list, out_dir, log_file, logname):
    """Function: split a FASTQ into a number of sub-files based on 
                 sequence barcode list
    This function calls ICELL8TCR_Demux.jar to demultiplex                    
    Parameters can be input as string.
    Returns True if successfull."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function split_fastq: Started for " + barcode_fastq \
              + " and " + mate_fastq, module_logname)

    split_fastq_cmd = "java -jar " + demux_tool_loc \
                      + " " + mate_fastq + " " + barcode_fastq + " " \
                      + " " + barcode_list + " " + out_dir

    logfile = log_file
    errfile = log_file.replace(".stdout", ".stderr")

    run_system_cmd(split_fastq_cmd, logfile, errfile,
                   "subprocess on split_fastq_cmd failed", module_logname,
                   append_mode=True)

    if os.path.splitext(barcode_list)[1] == ".csv":
        os.remove(barcode_list)
    if os.path.exists(os.path.splitext(barcode_list)[0].replace(
        "_printed","") + ".csv"):
        os.remove(os.path.splitext(barcode_list)[0].replace(
            "_printed","") + ".csv")
    else:
        pass

    write_log("info",
              "Function split_fastq for " + barcode_fastq + " and " + 
              mate_fastq + " Succesfully completed", module_logname)
    return True


def clean_intermediates(file_dir, logname):
    """Function: clean intermediate fastqs in a directory
    Returns True if successfull."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function clean_intermediates started",
              module_logname)

    for filename in glob.glob(os.path.join(file_dir, r"INVALID*.fastq")):
      try:
          os.remove(filename)
      except OSError as err:
          write_log("error", "Unable to clean files", module_logname)
          return False

    write_log("info",
              "Function clean_intermediates was succesfully completed",
              module_logname)

    return True


def add_info_to_file(filename, prefix, info, suffix, logname):
    """ Function: write info to stdout or stderr file
    This function is writen for MiXCR, 
    as MiXCR doesn't automatically write sample name in its stdout/stderr,
    prefix & suffix take a string as input
    """
    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info",
              "Function add_info_to_file: Started", module_logname)

    try:
        with open(filename, "a") as text_file:
            text_file.write(prefix + info + suffix)
    except EnvironmentError as err:
        write_log("error", "Unable to open file: " + text_file + "\n" +
                  str(err), module_logname)

    write_log("info", "Function add_info_to_file: Succesfully completed",
              module_logname)
    return True


def dos2unix(file,logname):
    """ Function: convert dos format file with \r\n to Unix with \n
    This function is writen for MiXCR, as MiXCR only use \n as EOL
    """
    try:
        text = open(file).read()
        open(file, 'w',newline='\n').write(text)
    except EnvironmentError as err:
        write_log("error", "Unable to open file: " + file + "\n" +
                  str(err), module_logname)
    return True


def mixcr_align(mixcr_loc, immu_loci, species,
                sample_r1_fastq, sample_r2_fastq,
                file_vdjca, align_report,
                stdout_file, stderr_file, logname):
    """ Function: mixcr aligns raw reads to reference
    V, D, J and C genes of T- and B- cell receptors.
    Returns True if successfull."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function mixcr_align: Started for " + 
              sample_r1_fastq + " and " + sample_r2_fastq, 
              module_logname)

    mixcr_align_cmd = "java -Xmx4g -Xms3g -jar " + mixcr_loc + " align " \
                      + " -c " + immu_loci + " -s " + species + " " \
                      + sample_r1_fastq + " " + sample_r2_fastq + " " \
                      + file_vdjca + " --report " + align_report

    run_system_cmd(mixcr_align_cmd, stdout_file, stderr_file,
                   "subprocess on mixcr_align_cmd failed", module_logname,
                   append_mode=True)

    write_log("info",
              "Function mixcr_align for " + sample_r1_fastq + " and " \
              + sample_r2_fastq + " Succesfully completed", module_logname)
    return True


def mixcr_assemble(mixcr_loc, file_vdjca, file_clns, clone_report,
                   stdout_file, stderr_file, logname):
    """ Function: mixcr assemble clones
    Returns True if successfull."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function mixcr_assemble: Started for "
              + file_vdjca, module_logname)

    mixcr_assemble_cmd = "java -Xmx4g -Xms3g -jar " + mixcr_loc \
                         + " assemble " + file_vdjca \
                         + " " + file_clns + \
                         " --report " + clone_report

    run_system_cmd(mixcr_assemble_cmd, stdout_file, stderr_file,
                   "subprocess on mixcr_assemble_cmd failed",
                   module_logname, append_mode=True)

    write_log("info",
              "Function mixcr_assemble for " + file_vdjca \
              + " Succesfully completed", module_logname)
    return True


def mixcr_export_clones(mixcr_loc, file_clns, file_clone_all,
                        stdout_file, stderr_file, logname):
    """ Function: mixcr export information about clones
    Returns True if successfull."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function mixcr_export_clones: Started for "
              + file_clns, module_logname)

    mixcr_export_clones_cmd = "java -Xmx4g -Xms3g -jar " + mixcr_loc \
                              + " exportClones " \
                              + " --preset full " + file_clns \
                              + " " + file_clone_all

    run_system_cmd(mixcr_export_clones_cmd, stdout_file, stderr_file,
                   "subprocess on mixcr_export_clones_cmd failed",
                   module_logname, append_mode=True)

    write_log("info",
              "Function mixcr_export_clones for " + file_clns \
              + " Succesfully completed", module_logname)
    return True


def write_summary_header(summary_file):
    """ Function: write header info of the summary.csv file
    """
    with open(summary_file, "w") as csv_file:
        header_barcode = "All_name,Sample," \
                         "Barcode," \
                         "Total_reads," \
                         "Aligned_reads," \
                         "Aligned (percent of total)," \
                         "Final clonotype count," \
                         "Reads_used_in_clonotypes," \
                         "Reads_used (percent of total)," \
                         "Top1_clonotype," \
                         "Top1_clonotype (fraction),Top2_clonotype," \
                         "Top2_clonotype (fraction),Top3_clonotype," \
                         "Top3_clonotype (fraction)," \
                         "Clonotype Count (>= 0.01%)," \
                         "Clonotype Count (>=0.01% and read 10)"
        header_clone = "Clone ID,Clone count," \
                       "Clone fraction,Clonal sequence(s)," \
                       "Clonal sequence quality(s)," \
                       "All V hits with score, All D hits with score," \
                       "All J hits with score, All C hits with score," \
                       "All V alignments,All D alignments," \
                       "All J alignments,N. Seq. CDR3,Min. qual. CDR3," \
                       "AA. Seq. CDR3,Ref. points"
        header_stats = "Alignment failed no hits (not TCR/IG?)," \
                       "Alignment failed because of absence of V hits," \
                       "Alignment failed because of absence of J hits," \
                       "Alignment failed because of low total score," \
                       "Overlapped," \
                       "Overlapped and aligned," \
                       "Overlapped and not aligned," \
                       "Reads used as core (percent of used)," \
                       "Mapped low quality reads (percent of used)," \
                       "Reads clustered in PCR error correction (" \
                       "percent of used)," \
                       "Clonotypes eliminated by PCR error correction," \
                       "Percent of reads dropped (lack of clonal sequence)," \
                       "Percent of reads dropped (low quality)," \
                       "Percent of reads dropped (failed mapping)," \
                       "Total V hits,Total J hits"
        csv_file.write(header_barcode + "," + header_clone + ","
                       + header_clone + "," + header_clone + ","
                       + header_stats + "\n")


def initiate_global_count():
    """ Function: initiate global variables, which would be used to count while
    iterate mixcr results for individual cells
    """
    global count_total
    global count_invalid
    global count_valid
    global count_selected
    global count_notselected
    global count_total_aligned
    global count_invalid_aligned
    global count_valid_aligned
    global count_selected_aligned
    global count_notselected_aligned
    count_total = int(0)
    count_invalid = int(0)
    count_valid = int(0)
    count_selected = int(0)
    count_notselected = int(0)
    count_total_aligned = int(0)
    count_invalid_aligned = int(0)
    count_valid_aligned = int(0)
    count_selected_aligned = int(0)
    count_notselected_aligned = int(0)


def summarize_mixcr_results_by_sample(
        analysis_dir, sample_name, report_name,
        align_subdir, assemble_subdir, all_clones_subdir):
    """ Function: summarize_mixcr_results_by_sample, 
    iterate individual result files of mixcr,
    and summarize statistics & clonotypes
    """
    global count_total
    global count_invalid
    global count_valid
    global count_selected
    global count_notselected
    global count_total_aligned
    global count_invalid_aligned
    global count_valid_aligned
    global count_selected_aligned
    global count_notselected_aligned

    # initialize variables
    align_report_line_num = int(0)
    desc_barcode = ""
    desc_stats = ""
    align_filter_short_dict = {}
    align_filter_full_dict = {
        "Alignment failed, no hits (not TCR/IG?)": "",
        "Alignment failed because of absence of V hits": "",
        "Alignment failed because of absence of J hits": "",
        "Alignment failed because of low total score": ""}
    clones_report_line_num = int(0)
    flag_clone = int
    desc_clone = ""
    all_clones_line_num = int(0)
    desc_v = ""
    desc_j = ""
    count_clone = int(0)
    count_clone_10 = int(0)
    count_a = int(0)
    count_b = int(0)
    count_3 = int(0)

    # set up file names for retriving data and save data
    file_align_report = os.path.join(analysis_dir, align_subdir,
                        sample_name + "_align_report.txt")
    file_clones_report = os.path.join(analysis_dir, assemble_subdir,
                         sample_name + "_clones_report.txt")
    file_clones_all = os.path.join(analysis_dir, all_clones_subdir,
                      sample_name + "_clones_all.txt")
    summary_file = os.path.join(analysis_dir, report_name + "_summary.csv")

    # set up barcode columns
    if sample_name != "INVALID-NA":
        desc_barcode = sample_name + "," + sample_name.split("-")[0] \
                       + "," + sample_name.split("-")[1]
    elif sample_name == "INVALID-NA":
        desc_barcode = sample_name + ",Invalid_Read,NA"

    # get data from xx_align_report.txt file
    with open(file_align_report) as align_report:
        for line in align_report:
            align_report_line_num += 1
            align_report_line_content = line.rstrip().split(":")

            if align_report_line_num == 7:
                desc_barcode += "," + align_report_line_content[1]
                count_total += int(align_report_line_content[1])
                if sample_name == "INVALID-NA":
                    count_invalid = int(align_report_line_content[1])
                else:
                    count_valid += int(align_report_line_content[1])
                    if sample_name[0:11] == "notselected":
                        count_notselected += int(align_report_line_content[1])
                    else:
                        count_selected += int(align_report_line_content[1])

            if align_report_line_num == 8:
                desc_barcode += "," + align_report_line_content[1].\
                    split("(")[0]
                desc_barcode += "," + align_report_line_content[1].\
                    split("(")[1].split(")")[0]
                count_total_aligned += int(align_report_line_content[1]
                                           .split("(")[0])
                if sample_name == "INVALID-NA":
                    count_invalid_aligned = int(align_report_line_content[1]
                                                .split("(")[0])
                else:
                    count_valid_aligned += int(align_report_line_content[1]
                                               .split("(")[0])
                    if sample_name[0:11] == "notselected":
                        count_notselected_aligned += int(
                            align_report_line_content[1].split("(")[0])
                    else:
                        count_selected_aligned += int(
                            align_report_line_content[1].split("(")[0])

            if align_report_line_num in list(range(9, 13)):
                if align_report_line_content[0] in \
                        ["Alignment failed, no hits (not TCR/IG?)",
                         "Alignment failed because of absence of V hits",
                         "Alignment failed because of absence of J hits",
                         "Alignment failed because of low total score"]:
                    filter_key, filter_value = \
                        align_report_line_content[0], \
                        align_report_line_content[1]
                    align_filter_short_dict[filter_key] = filter_value

    align_filter_full_dict.update(align_filter_short_dict)

    if len(align_filter_full_dict[
               "Alignment failed, no hits (not TCR/IG?)"]) > 0:
        desc_stats += "," + align_filter_full_dict[
            "Alignment failed, no hits (not TCR/IG?)"].\
            split("(")[1].split(")")[0]
    else:
        desc_stats += ",NA"

    if len(align_filter_full_dict[
               "Alignment failed because of absence of V hits"]) > 0:
        desc_stats += "," + align_filter_full_dict[
            "Alignment failed because of absence of V hits"].\
            split("(")[1].split(")")[0]
    else:
        desc_stats += ",NA"

    if len(align_filter_full_dict[
               "Alignment failed because of absence of J hits"]) > 0:
        desc_stats += "," + align_filter_full_dict[
            "Alignment failed because of absence of J hits"].\
            split("(")[1].split(")")[0]
    else:
        desc_stats += ",NA"

    if len(align_filter_full_dict[
               "Alignment failed because of low total score"]) > 0:
        desc_stats += "," + align_filter_full_dict[
            "Alignment failed because of low total score"].\
            split("(")[1].split(")")[0]
    else:
        desc_stats += ",NA"

    with open(file_align_report) as align_report:
        for line in align_report:
            align_report_line_content = line.rstrip().split(":")
            if align_report_line_content[0] in ["Overlapped",
                                                "Overlapped and aligned",
                                                "Overlapped and not aligned"]:
                desc_stats += "," + align_report_line_content[1].split(
                    "(")[1].split(")")[0]

    # get data from xx_clones_report.txt file
    with open(file_clones_report) as clones_report:
        for line in clones_report:
            clones_report_line_num += 1
            clones_report_line_content = line.rstrip().split(":")
            if clones_report_line_num == 7:
                desc_barcode += "," + clones_report_line_content[1]
                # check if final clonotype count is zero
                if int(clones_report_line_content[1]) == 0:
                    flag_clone = int(1)
                elif int(clones_report_line_content[1]) > 0:
                    flag_clone = int(0)
                else:
                    print("clone number cannot be negative")
                    break
            if clones_report_line_num == 9:
                desc_barcode += "," + clones_report_line_content[1].\
                    split("(")[0]
                desc_barcode += "," + clones_report_line_content[1].\
                    split("(")[1].split(")")[0]
            if clones_report_line_num in list(range(11, 14)):
                desc_stats += "," + clones_report_line_content[1].\
                    split("(")[1].split(")")[0]
            if clones_report_line_num == 19:
                desc_stats += "," + clones_report_line_content[1]

    with open(file_clones_report) as clones_report:
        for line in clones_report:
            clones_report_line_content = line.rstrip().split(":")
            if clones_report_line_content[0] in [
                "Reads dropped due to the lack of a clone sequence",
                "Reads dropped due to low quality",
                "Reads dropped due to failed mapping"]:
                desc_stats += "," + clones_report_line_content[1].\
                    split("(")[1].split(")")[0]

    # if final clonotype count is zero, then all barcode clone and stats are NA
    if flag_clone == 1:
        desc_barcode += ",NA,NA,NA,NA,NA,NA"
        # desc_clone for clone_A, clone_B and clone_3 are all NAs
        desc_clone += \
            ",NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"
        desc_clone += \
            ",NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"
        desc_clone += \
            ",NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"
        desc_stats += ",NA,NA"

    # if final clonotype count > 0,
    #   open file_clones_all and summarize info of all clones
    elif flag_clone == 0:
        # get data from xx_clone_all.txt file
        with open(file_clones_all) as all_clones:
            for line in all_clones:
                all_clones_line_num += 1
                all_clones_line_content = line.rstrip()
                all_clones_line_content = \
                    re.split(r'\t', all_clones_line_content.rstrip('\t'))
                all_clones_line_content[5] = \
                    re.sub(r",", "&", all_clones_line_content[5])
                all_clones_line_content[7] = \
                    re.sub(r",", "&", all_clones_line_content[7])
                all_clones_line_content[8] = \
                    re.sub(r",", "&", all_clones_line_content[8])
                if len(all_clones_line_content[6]) == 0:
                    temp1 = "NA"
                elif len(all_clones_line_content[6]) > 0:
                    all_clones_line_content[6] = \
                        re.sub(r",", "&", all_clones_line_content[6])
                    temp1 = all_clones_line_content[6]
                if len(all_clones_line_content[10]) == 0:
                    temp2 = "NA"
                elif len(all_clones_line_content[10]) > 0:
                    temp2 = all_clones_line_content[10]

                # get clonotypes
                if all_clones_line_content[5] != "":
                    g = all_clones_line_content[5].split("*")
                    desc_v += g[0] + ":" + all_clones_line_content[1] + ":" + \
                              all_clones_line_content[2] + "&"
                if all_clones_line_content[7] != "":
                    h = all_clones_line_content[7].split("*")
                    desc_j += h[0] + ":" + all_clones_line_content[1] + ":" + \
                              all_clones_line_content[2] + "&"

                # get top a/b clonotypes
                # look only the first three lines: 2, 3, 4, 
                #  and check first three clones
                if all_clones_line_num in list(range(2,5)):    
                    # count total clone number
                    if float(all_clones_line_content[2]) >= 0.0001:
                        count_clone += 1
                        if int(all_clones_line_content[1]) >= 10:
                            count_clone_10 += 1  
                            # count clones with more than 10 counts                       
                    if all_clones_line_content[5][2] == "A":
                        # if clone type is A
                        desc_barcode += ",TCRa," + all_clones_line_content[2]
                        # mark clone type and fraction
                        count_a += 1
                        count_3 += 1 
                        # count_3 record how many clonotype found 
                        # within the top3 clonotypes
                    elif all_clones_line_content[5][2] == "B":
                        desc_barcode += ",TCRb," + all_clones_line_content[2]
                        count_b += 1
                        count_3 += 1
                    # append info of current clone to desc_clone
                    # append clone in first line
                    desc_clone += "," + all_clones_line_content[0] + \
                                  "," + all_clones_line_content[1] + \
                                  "," + all_clones_line_content[2] + \
                                  "," + all_clones_line_content[3] + \
                                  "," + all_clones_line_content[4] + \
                                  "," + all_clones_line_content[5] + \
                                  "," + temp1 + \
                                  "," + all_clones_line_content[7] + \
                                  "," + all_clones_line_content[8] + \
                                  "," + all_clones_line_content[9] + \
                                  "," + temp2 + \
                                  "," + all_clones_line_content[11] + \
                                  "," + all_clones_line_content[23] + \
                                  "," + all_clones_line_content[24] + \
                                  "," + all_clones_line_content[32] + \
                                  "," + all_clones_line_content[34]
                 
                
    # after nagvigate the first three line, if count_3=1, only one clone foun
    if count_3 == 1:
        desc_barcode += ",NA,NA,NA,NA"
        desc_clone += ",NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"
        desc_clone += ",NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"
    if count_3 == 2: 
        desc_barcode += ",NA,NA"                
        desc_clone += ",NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"   
               
    desc_barcode += "," + str(count_clone) + "," + str(count_clone_10)
    desc_stats += "," + desc_v + "," + desc_j

    temp_desc = desc_barcode + desc_clone + desc_stats + "\n"
    
    if len(temp_desc) <= 32767:
        with open(summary_file, "a") as csv_file:
            csv_file.write(temp_desc)
    else:
        desc_list = [temp_desc[i: i + (32766-80)] for i in
                     range(0, len(temp_desc), (32766-80))]
        num_desc = len(desc_list)
        with open(summary_file, "a") as csv_file:
                csv_file.write( desc_list[0] + "\n")
                for str_idx in range(1,num_desc):
                    comma_str = ''.join(desc_list[0:str_idx])
                    num_comma = comma_str.count(",")
                    str2write = ''.join(","*num_comma) + desc_list[str_idx] + \
                                "\n"
                    csv_file.write(str2write)


def vdjviz_convert(file_clones_all, chord_file):
    """ Function: vdjviz_convert, convert mixcr clones_all.txt
        into vdjviz input file
    """
    if os.path.getsize(file_clones_all) > 0:
        with open(chord_file, "w") as out_file:
            headers = ['#count','freq','cdr3nt','cdr3aa','v','d','j']
            out_file.write("\t".join(headers)+"\n")
        with open(file_clones_all) as all_clones:
                all_clones_line_num = 0
                for line in all_clones:
                    all_clones_line_num += 1
                    if all_clones_line_num != 1:
                        all_clones_line_content = line.rstrip()
                        all_clones_line_content = \
                            re.split(r"\t", 
                                all_clones_line_content.rstrip("\t"))
                        all_clones_line_content[5] = \
                            all_clones_line_content[5].split("*")[0]
                        all_clones_line_content[6] = \
                            all_clones_line_content[6].split("*")[0]
                        all_clones_line_content[7] = \
                            all_clones_line_content[7].split("*")[0]
                        if len(all_clones_line_content[5]) == 0:
                            all_clones_line_content[5]=str(r".")
                        if len(all_clones_line_content[6]) == 0:
                            all_clones_line_content[6]=str(r".")
                        if len(all_clones_line_content[7]) == 0:
                            all_clones_line_content[7]=str(r".") 
                        newline = [str(all_clones_line_content[1]),
                                   str(all_clones_line_content[2]),
                                   str(all_clones_line_content[23]),
                                   str(all_clones_line_content[32]),
                                   str(all_clones_line_content[5]),
                                   str(all_clones_line_content[6]), 
                                   str(all_clones_line_content[7])]
                        with open(chord_file, "a") as out_file:
                            out_file.write("\t".join(newline[0:])+"\n")
                if all_clones_line_num < 3:
                    os.remove(chord_file)
                else:
                    pass
    else:
        pass


def count_read_in_fastq(analysis_dir,subdir,barcode_dict,read1_fastq,
                        report_name):
    """ Function: count_read_in_fastq, iterate individual cell fastq, 
    count total reads"""
    desc_reads = ""
    for key, value in barcode_dict.items():
        if key == "INVALID":
            continue
        else:
            fastq = os.path.join(analysis_dir, subdir, key + "_"
                    + os.path.basename(read1_fastq).replace(".gz",""))
            num_lines = int(sum(1 for line in open(fastq)))
            num_reads = int(num_lines/4)
            desc_reads += key + "," + value + "," + str(num_reads) + "\n"
     
    read_counts_file = os.path.join(analysis_dir, report_name +
                           "_reads_print.csv")

    with open(read_counts_file, "w") as csv_file:
            csv_file.write("Sample_Type,Barcode,Total_Read\n") 
            csv_file.write(desc_reads)


def count_barcode_in_fastq(barcode_fastq, outdir, outfile_prefix):
    """ Function: count_barcode_in_fastq, count reads number for all unique
    barcodes in current fastq, particularly invalid fastq"""
    i=0 # initiate line number
    fastq_barcode_dict = {} # initiate invalid barcode dictionary
    # get data from barcode_fastq
    with open(barcode_fastq) as fastq:
        for line in fastq:
            i += 1 # count line number
            j = i%4 # get remainder

            if j == 2:
                barcode = line.rstrip()[0:10]
                if barcode in fastq_barcode_dict.keys():
                    fastq_barcode_dict[barcode] += 1
                else:
                    fastq_barcode_dict[barcode] = 1
            else:
                pass

    # initiate outfile
    outfile = os.path.join(outdir, outfile_prefix + 
              "_reads_invalid.csv")

    # write header of invalid barcode count
    with open(outfile,"w") as csv_file:
        csv_file.write("Invalid_Barcode,Total_Read\n")

    # write invalid barcode_dict to outfile as csv format
    with open(outfile, "a") as invalid_barcode_count:
        writer01 = csv.writer(invalid_barcode_count)
        for row in fastq_barcode_dict.items():
            writer01.writerow(row)


def write_all_samples_stats(analysis_dir, report_name):
    """Function: write_all_samples_stats
    write all sample statistics"""
    global count_total
    global count_invalid
    global count_valid
    global count_selected
    global count_notselected
    global count_total_aligned
    global count_invalid_aligned
    global count_valid_aligned
    global count_selected_aligned
    global count_notselected_aligned
    summary_file = os.path.join(analysis_dir, report_name + "_summary.csv")
    with open(summary_file, "a") as csv_file:
        csv_file.write("Sample Name," + report_name + "\n")
        csv_file.write("Total Reads," + str(count_total) + "\n")
        csv_file.write("Invalid Barcode Reads," + str(count_invalid) + "\n")
        csv_file.write("Printed Barcode Reads," + str(count_valid) + "\n")
        csv_file.write("Selected Barcode Reads," + str(count_selected) + "\n")
        csv_file.write("NotSelected Barcode Reads,"
                       + str(count_notselected) + "\n")
        csv_file.write("Total Aligned Reads," +
                       str(count_total_aligned) + "\n")
        csv_file.write("Invalid Barcode Aligned Reads," +
                       str(count_invalid_aligned) + "\n")
        csv_file.write("Printed Barcode Aligned Reads," +
                       str(count_valid_aligned) + "\n")
        csv_file.write("Selected Barcode Aligned Reads,"
                       + str(count_selected_aligned) + "\n")
        csv_file.write("NotSelected Barcode Aligned Reads,"
                       + str(count_notselected_aligned) + "\n")


def write_all_samples_brief_stats(analysis_dir, report_name):
    """Function: write_all_samples_brief_stats
    write brief statstics"""
    global count_selected
    global count_selected_aligned
    summary_file = os.path.join(analysis_dir, report_name + "_summary.csv")
    with open(summary_file, "a") as csv_file:
        csv_file.write("Sample Name," + report_name + "\n")
        csv_file.write("Selected Barcode Reads," + str(count_selected) + "\n")
        csv_file.write("Selected Barcode Aligned Reads," +
                       str(count_selected_aligned) + "\n")


def do_tertiary(file_dir, report_name, neg_ctrl_name, logname):
    """Function do_tertiary: apply tertiary analysis to summmary.csv
    Return True if successful."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function do_tertiary: Started",
              module_logname)

    summary_file = report_name + "_summary.csv"

    with open(os.path.join(file_dir, summary_file), 'r') as f:
        lines = f.readlines()

    with open(os.path.join(file_dir, summary_file), 'w') as fout:
        for line in lines:
            fout.writelines(re.sub(r",+$", "", line))

    summ = pd.read_csv(os.path.join(file_dir, summary_file))
    file_prefix = report_name

    # --------------------------------------------
    #   exclude summary stats, invalid and notselected
    # --------------------------------------------
    # exclude summary stats
    if any(summ. All_name =="Sample Name"):
        ind = summ.index[summ. All_name =="Sample Name"].tolist()[0]
        summ = summ.loc[:( ind -1), :]
        del ind
        # print(summ.shape)

    # exclude invalid stats
    if any(summ. All_name =="INVALID-NA"):
        ind = summ.index[summ. All_name =="INVALID-NA"].tolist()[0]
        summ = summ.loc[:( ind -1) ,:]
        del ind

    # excluded notselected
    if summ.All_name.str.contains('notselected').any():
        ind = summ.index[summ.All_name.str.contains('notselected')]
        summ = summ.drop(summ.index[ind])
        del ind

    # --------------------------------------------
    #   summarize cells & their sample number
    # --------------------------------------------
    # if sample named as celltype_C[0-9]_R[0-9] format, substract cell type
    # if no such format, use sample name as cell type
    pd.options.mode.chained_assignment = None
    summ.loc[:,'Type'] = summ.Sample.apply(
        lambda x: x[0:re.search("_R[0-9]+_C[0-9]+", x).span()[0]] if 
        bool(re.search("_R[0-9]+_C[0-9]+", x)) else x)

    # create another dataframe to store filtering stats
    stats = summ

    # -----------------
    #     filters
    # -----------------
    # 1) clonotype NA filter
    # if all clonotypes are NA, filter out
    # exclude such samples except for negative controls
    # if any negative controls with no clonotype found, 
    #   set NA in top1 clono count to 0,
    #   as such data point is great data point, not missing data

    # find if negative control string is in dataset, if not, raise warning
    if summ.Type.str.contains(neg_ctrl_name).any():
        pass
    elif summ.Type.str.contains(neg_ctrl_name).sum() == 0:
        print(
          "The negative control name specified was not found in sample.\n"
          "Count filter based on negative control was not used.")

    # find index with all three top clonotypes as null
    clone_col = [col for col in summ.columns if 
                 bool(re.search("^Top.*clonotype$", col))]
    na_ind = summ.index[summ[clone_col].isnull().all(axis=1) 
                        == True].tolist()
    neg_ind = summ.index[summ.Type.str.contains(neg_ctrl_name) 
                         == True].tolist()
    ex_ind = list(set(na_ind ) -set(neg_ind))
    if len(na_ind ) >0:
        summ = summ.drop(ex_ind, axis=0)
    del clone_col, na_ind, neg_ind, ex_ind

    # summarize NA stats in the dataframe named as stats
    stats['top1_na'] = stats.Top1_clonotype.isnull()
    stats['top2_na'] = stats.Top2_clonotype.isnull()
    stats['top3_na'] = stats.Top3_clonotype.isnull()
    clone_col = [col for col in stats.columns 
                 if bool(re.search("^Top.*clonotype$", col))]
    stats['all_na_rd1'] = stats[clone_col].isnull().all(axis=1).tolist()
    del clone_col

    # if any negative controls with no clonotype found, 
    #  set NA in top1 clono count to 0,
    #  as such data point is great data point, not missing data
    neg_na_ind = np.where(summ['Clone count'].isnull() & 
                          summ['Sample'].str.contains(neg_ctrl_name))[0].tolist()
    if len(neg_na_ind) > 0:
        pd.options.mode.chained_assignment = None
        summ.iloc[neg_na_ind, 18] = 0
    del neg_na_ind

    # 2) filter based on mean+3sd of negative controls' top1 clonotype count
    # use Top1 clonotype's count as cutoff for all three clonotype,
    # any clonotype filtered out, set corresponding its relavent info to NA
    neg_ind = np.where(summ['Sample'].str.contains(neg_ctrl_name))[0].tolist()
    clone1_count_ind = np.where(summ.columns.str.contains("Clone count")
                                )[0].tolist()[0]
    count_filter = np.ceil(np.mean(summ.iloc[neg_ind, clone1_count_ind]) +
                         3*np.std(summ.iloc[neg_ind, clone1_count_ind]))
    # if count_filter is nan - no negative control in data set, reset it to 0
    if np.isnan(count_filter):
        count_filter = 0
    del neg_ind, clone1_count_ind

    # reset clonotypes whose count number are less than count_filter to NaN
    for count_column in ['Clone count', 'Clone count.1', 'Clone count.2']:
        count_list = [sys.maxsize if pd.isnull(x) else x for x in 
                      summ.loc[:, count_column]]
        boolean_list = [True if x < count_filter else False 
                        for x in count_list]
        del_ind_list = [i for i, x in enumerate(boolean_list) if x]
        if len(del_ind_list) > 0:
            count_ind = summ.columns.get_loc(count_column)
            for j in del_ind_list:
                if count_ind == 18:
                    summ.iloc[j, np.r_[9:11,17:33]] = float("nan")
                elif count_ind == 34:
                    summ.iloc[j, np.r_[11:13, 33:49]] = float("nan")
                elif count_ind == 50:
                    summ.iloc[j, np.r_[13:15, 49:65]] = float('nan')
            del count_ind, j
    del count_column, count_list, boolean_list, del_ind_list

    # summarize stats in stats data frame: if count less than threshold, 
    #   assign corresponding clonotype as nan
    n = 0
    for count_column in ['Clone count', 'Clone count.1', 'Clone count.2']:
        n += 1
        count_list = [sys.maxsize if pd.isnull(x) else x 
                      for x in stats.loc[:, count_column]]
        stats["count_drop_clono" + str(n)] = [True if x < count_filter 
                                              else False for x in count_list]
        if len([i for i, x in 
                enumerate(stats["count_drop_clono" + str(n)]) if x]) > 0:
            for j in [i for i, x in 
                      enumerate(stats["count_drop_clono" + str(n)]) if x]:
                stats.iloc[j, (7+2*n)] = float("nan")
            del j
    del n, count_list, count_column

    # after above steps, if there are samples with all clonotypes as NA,
    #   drop them (including negative controls)
    # find index with all three top clonotypes as null
    clone_col = [col for col in summ.columns 
                 if bool(re.search("^Top.*clonotype$", col))]
    na_ind = summ.index[summ[clone_col].isnull().all(axis=1) == True].tolist()
    if len(na_ind) > 0:
        summ = summ.drop(na_ind, axis=0)
    del clone_col, na_ind

    # record filtering stats in stats data frame
    clone_col = [col for col in stats.columns 
                 if bool(re.search("^Top.*clonotype$", col))]
    stats['all_na_rd2'] = stats[clone_col].isnull().all(axis=1).tolist()
    del clone_col


    # 3) CDR3 deletion filter
    # if any deletion found in AA sequence, 
    #    set all info of corresponding clonotype to NA
    for aa_seq_column in ['AA. Seq. CDR3','AA. Seq. CDR3.1','AA. Seq. CDR3.2']:
        null_n_seq_list = ['null' if pd.isnull(x) else x 
                           for x in summ.loc[:, aa_seq_column]]
        boolean_list = [True if ("_" in x or "*" in x)  
                        else False for x in null_n_seq_list]
        del_ind_list = [i for i, x in enumerate(boolean_list) if x]
        if len(del_ind_list) > 0:
            aa_seq_ind = summ.columns.get_loc(aa_seq_column)
            for j in del_ind_list:
                if aa_seq_ind == 31:
                    summ.iloc[j, np.r_[9:11,17:33]] = float("nan")
                elif aa_seq_ind == 47:
                    summ.iloc[j, np.r_[11:13, 33:49]] = float("nan")
                elif aa_seq_ind == 63:
                    summ.iloc[j, np.r_[13:15, 49:65]] = float('nan')
            del aa_seq_ind, j
    del aa_seq_column, null_n_seq_list, boolean_list, del_ind_list

    # summarize stats in stats dataframe: if deletion or frameshift 
    #   found in AA sequence, set clonotype as nan
    n = 0
    for aa_seq_column in ['AA. Seq. CDR3','AA. Seq. CDR3.1','AA. Seq. CDR3.2']:
        n += 1
        null_n_seq_list = ['null' if pd.isnull(x) else x 
                           for x in stats.loc[:, aa_seq_column]]
        stats["AA_deletion_drop_clono" + str(n)] = [True 
            if ("_" in x or "*" in x) else False for x in null_n_seq_list]
        if len([i for i, x in 
               enumerate(stats["AA_deletion_drop_clono" + str(n)]) if x]) > 0:
            for j in [i for i, x in 
                enumerate(stats["AA_deletion_drop_clono" + str(n)]) if x]:
                stats.iloc[j, (7+2*n)] = float("nan")
            del j
    del n, null_n_seq_list, aa_seq_column

    # after drop AA deletion in CDR3, 
    # if there are samples with all clonotypes as NA,
    #   drop them (including negative controls)
    clone_col = [col for col in summ.columns 
                 if bool(re.search("^Top.*clonotype$", col))]
    na_ind = summ.index[summ[clone_col].isnull().all(axis=1) == True].tolist()
    if len(na_ind) > 0:
        summ = summ.drop(na_ind, axis=0)
    del clone_col, na_ind

    # record filtering stats in stats dataframe
    clone_col = [col for col in stats.columns 
                 if bool(re.search("^Top.*clonotype$", col))]
    stats['all_na_rd3'] = stats[clone_col].isnull().all(axis=1).tolist()
    del clone_col

    # format stats dataframe and generate summary tables
    stats = stats.iloc[:, np.r_[1:5, 7, 6, 81:94]]
    stats.loc[:, "number_cell"] = np.repeat(1, len(stats))

    # Table1.Number of Remaining Cells by filters
    table1 = stats.groupby(['Type']).sum().loc[:, 
        ['number_cell', 'all_na_rd1', 'all_na_rd2', 'all_na_rd3']].astype(int)
    for n in list(range(1,4)):
        table1["No.cell_after_filter" + str(n)] = \
        table1['number_cell'] - table1.iloc[:,n]
    del n
    table1 = table1.iloc[:, np.r_[0, 4:7]]

    # write summary tables to cell_filter_tables.csv
    with open(os.path.join(file_dir, file_prefix + 
              "_cell_filter_tables.csv"),
              "w") as filter_table_file:
        filter_table_file.write(
            "Table1.Number of Remaining Cells by filters\n")
        table1.to_csv(filter_table_file)

    # Table2. Percentage Remaining by filters
    table2 = table1.div(table1.number_cell, axis='index')
    for col in list(range(0, 4)):
        table2.iloc[:,col] = pd.Series(["{0:.2f}%".format(val * 100)
                                                for val in table2.iloc[:,col]],
                                                           index=table2.index)
    with open(os.path.join(file_dir, file_prefix + "_cell_filter_tables.csv"),
              "a") as filter_table_file:
        filter_table_file.write("Table2. Percentage Remaining by filters\n")
        table2.to_csv(filter_table_file)

    # Table3. Detailed Filtering Stats
    table3 = stats.groupby(['Type']).sum().loc[:, [
              "number_cell","top1_na","top2_na", "top3_na", "all_na_rd1",
              "count_drop_clono1", "count_drop_clono2", 
              "count_drop_clono3", "all_na_rd2", 
              "AA_deletion_drop_clono1", "AA_deletion_drop_clono2", 
              "AA_deletion_drop_clono3", 
              "all_na_rd3"]].astype(int)

    table3.iloc[:, 12] = table3.iloc[:,12] - table3.iloc[:, 8]
    table3.iloc[:, 8] = table3.iloc[:, 8] - table3.iloc[:, 4]
    table3.rename(columns={"all_na_rd1": "drop_filter1", 
                           "all_na_rd2": "drop_filter2", 
                           "all_na_rd3":"drop_filter3"}, 
                  inplace=True)

    with open(os.path.join(file_dir, file_prefix + "_cell_filter_tables.csv"),
              "a") as filter_table_file:
        filter_table_file.write("Table3. Detailed Filtering Stats\n")
        table3.to_csv(filter_table_file)


    # ----------------------------------------
    #    prepare simplified table
    # ----------------------------------------

    # after QC, check if all cells are filtered out
    if len(summ) == 0:
        print("No data is remaining after QC, \
          please check cell filter table for details.")
        write_log("info","Function do_tertiary in directory " + file_dir +
                  " Succesfully completed", module_logname)
        return True

    # subtract several columns from raw table
    dat = summ.iloc[:,np.r_[1:3, 3:5, 7, 6, 9:11, 18, 29, 31, 22:25, 11:13, 
                            34, 45, 47, 38:41, 13:15, 50, 61, 63, 54:57]]

    # summarize chain composition
    clone_col = [col for col in dat.columns 
                 if bool(re.search("^Top.*clonotype$", col))]
    clone_str_max_len = sorted(list(set([len(x) for x 
                            in dat[clone_col].fillna('').values.flatten()])),
                            reverse=True)[0]
    n = clone_str_max_len -1
    df = dat[clone_col].fillna('').applymap(lambda x: x[n:])
    df2 = df[['Top1_clonotype', 'Top2_clonotype', 'Top3_clonotype']]
    dat.loc[:, 'Chains'] = df2.apply(lambda x: ''.join(sorted(x)), axis=1)
    del clone_col, df, df2

    # summarize cell type
    dat.loc[:, 'Type'] = dat.Sample.apply(
        lambda x: x[0:re.search("_R[0-9]+_C[0-9]+", x).span()[0]] 
                    if bool(re.search("_R[0-9]+_C[0-9]+", x)) else x)

    # summarize trueChains, which will be used for TCRa & TCRb subset in the next section
    clone_col = [col for col in dat.columns 
                 if bool(re.search("^Top.*clonotype$", col))]
    clone_str_max_len = sorted(list(set([len(x) for x in 
                            dat[clone_col].fillna('').values.flatten()])), 
                            reverse=True)[0]
    n = clone_str_max_len -1
    fill_str = ''.join(list(np.repeat('0', n+1)))
    df = dat[clone_col].fillna(fill_str).applymap(lambda x: x[n:])
    df2 =  df[['Top1_clonotype', 'Top2_clonotype', 'Top3_clonotype']]
    dat.loc[:, 'TrueChains'] = df2.apply(lambda x: ''.join(x), axis=1)
    del clone_col, df, df2

    # find chain composition and order dat by cell type & chain composition
    chains_comp_list = sorted(list(set(dat['Chains'])))
    dat.loc[:, 'Chains'] = pd.Categorical(dat['Chains'], chains_comp_list)
    del chains_comp_list
    dat = dat.sort_values(['Type', 'Chains'])
    dat = dat.iloc[:,np.r_[0:2, 31, 30, 2:30, 32]]
    dat.columns = ["Sample","Barcode","Type","Chains",
                      "Total_Reads","Aligned_Reads",
                      "Reads_Used_in_Clonotypes","Total_Clonotypes",
                      "Clono1","Clono1_Freq","Clono1_Count",
                      "Clono1_N_Seq", "Clono1_AA_Seq",
                      "Clono1_V", "Clono1_D","Clono1_J",
                      "Clono2","Clono2_Freq","Clono2_Count",
                      "Clono2_N_Seq", "Clono2_AA_Seq",
                      "Clono2_V", "Clono2_D","Clono2_J",
                      "Clono3","Clono3_Freq","Clono3_Count",
                      "Clono3_N_Seq", "Clono3_AA_Seq",
                      "Clono3_V", "Clono3_D","Clono3_J","TrueChains"]

    # simplify V,D,J results
    # now, modify "Clono1_V", "Clono1_D","Clono1_J" and "Clono2_V", "Clono2_D","Clono2_J",
    # so that only single gene allele reported
    if not dat.Clono1_V.isnull().all():
        dat.Clono1_V = \
        dat.Clono1_V.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono1_D.isnull().all():
        dat.Clono1_D = \
        dat.Clono1_D.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono1_J.isnull().all():
        dat.Clono1_J = \
        dat.Clono1_J.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono2_V.isnull().all():
        dat.Clono2_V = \
        dat.Clono2_V.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono2_D.isnull().all():
        dat.Clono2_D = \
        dat.Clono2_D.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono2_J.isnull().all():
        dat.Clono2_J = \
        dat.Clono2_J.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono3_V.isnull().all():
        dat.Clono3_V = \
        dat.Clono3_V.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono3_D.isnull().all():
        dat.Clono3_D = \
        dat.Clono3_D.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)
    if not dat.Clono3_J.isnull().all():
        dat.Clono3_J = \
        dat.Clono3_J.apply(lambda x: 
          x[0:re.search("\\*", x).span()[0]] if isinstance(x, str) else x)

    # write simplified top3 clonotypes to csv
    dat.iloc[:, 0:33].to_csv(os.path.join(file_dir,file_prefix + 
        "_simplified_top3.csv"), index=False)

    # Table4. Sample Number by Alpha Beta Chain Composition
    table4 = pd.crosstab(dat['Type'], dat['Chains'])
    with open(os.path.join(file_dir, file_prefix + 
        "_chain_composition_tables.csv"),
              "w") as chain_table_file:
        chain_table_file.write(
          "Table4. Sample Number by Alpha Beta Chain Composition\n")
        table4.to_csv(chain_table_file)

    # Table5. Sample Fractions by Alpha Beta Chain Composition
    table5 = pd.crosstab(dat['Type'], dat['Chains']).\
        div(pd.crosstab(dat['Type'], dat['Chains']).sum(axis=1), axis='rows')
    for col in list(range(0, table5.shape[1])):
        table5.iloc[:,col] = \
            pd.Series(["{0:.2f}%".format(val * 100) 
              for val in table5.iloc[:,col]],index=table5.index)

    with open(os.path.join(file_dir, file_prefix + 
        "_chain_composition_tables.csv"),
              "a") as chain_table_file:
        chain_table_file.write(
          "Table5. Sample Fractions by Alpha Beta Chain Composition\n")
        table5.to_csv(chain_table_file)

    # --------------------------------------------------------
    #  summarize cell types, clonotypes & chains
    # --------------------------------------------------------

    # subset data set by cell type
    cell_type_list = list(set(dat.Type.unique().tolist()))
    cell_dataset_list = []
    for i in range(0, len(cell_type_list)):
      vars()["dat_"+cell_type_list[i]] = dat.loc[
          dat.Type == cell_type_list[i], :]
      cell_dataset_list.append(vars()["dat_" + cell_type_list[i]])

    # generate vdjviz input for cell_type & clonotype
    for dataset in cell_dataset_list:
        cell_type = dataset.Type.unique().tolist()
        tmp_df1 = dataset.iloc[:, np.r_[8,10:16]]
        tmp_df2 = dataset.iloc[:, np.r_[16, 18:24]]
        tmp_df3 = dataset.iloc[:, np.r_[24, 26:32]]
        tmp_df1.columns.values[:] = \
            ["clono", "count", "cdr3nt", "cdr3aa", "v", "d", "j"]
        tmp_df2.columns.values[:] = \
            ["clono", "count", "cdr3nt", "cdr3aa", "v", "d", "j"]
        tmp_df3.columns.values[:] = \
            ["clono", "count", "cdr3nt", "cdr3aa", "v", "d", "j"]
        tmp_df = tmp_df1.append(tmp_df2).append(tmp_df3)
        clono_type_list = [x for x in 
            tmp_df.clono.unique().tolist() if str(x) != 'nan']
        for j in range(0, len(clono_type_list)):
            temp_df = tmp_df.loc[tmp_df.clono == 
                clono_type_list[j], :].drop(['clono'], axis=1)
            if len(temp_df) > 0:
                temp_df.loc[:, 'nt_aa_vdj'] = \
                    temp_df['cdr3nt'].astype(str) + "&" + \
                    temp_df['cdr3aa'].astype(str) + "&" + \
                    temp_df['v'].astype(str) + "&" + \
                    temp_df['d'].astype(str) + "&" + \
                    temp_df['j'].astype(str)
                temp_agg = temp_df.groupby('nt_aa_vdj').\
                  agg({'count': sum}).reset_index()
                temp_agg = temp_agg.iloc[:, np.r_[1, 0]]
                temp_vdjviz = pd.concat([temp_agg['count'], 
                  temp_agg['nt_aa_vdj'].str.split('&', expand=True)], axis=1)
                temp_vdjviz.columns = \
                  ["#count", "cdr3nt", "cdr3aa", "v", "d", "j"]
                temp_vdjviz.loc[:, 'freq'] = \
                  temp_vdjviz['#count'] / sum(temp_vdjviz['#count'])
                temp_vdjviz = temp_vdjviz.iloc[:, np.r_[0, 6, 1:6]]
                temp_vdjviz = temp_vdjviz.replace('nan', '.', regex=True)
                temp_vdjviz["#count"] = temp_vdjviz["#count"].astype(np.int64)
                if temp_vdjviz.shape[0] < 2:               
                     pass
                else:
                    temp_vdjviz.to_csv(os.path.join(file_dir, cell_type[0] 
                                       + "_" + clono_type_list[j] + 
                                       "_clono_vdjviz.txt"),
                                       header=True, index=None, sep='\t')

    # generate cell count for vj join
    for dataset in cell_dataset_list:
        cell_type = dataset.Type.unique().tolist()
        tmp_df1 = dataset.iloc[:, np.r_[8, 13, 15]]
        tmp_df2 = dataset.iloc[:, np.r_[16, 21, 23]]
        tmp_df3 = dataset.iloc[:, np.r_[24, 29, 31]]
        tmp_df1.columns.values[:] = ["clono", "v",  "j"]
        tmp_df2.columns.values[:] = ["clono", "v",  "j"]
        tmp_df3.columns.values[:] = ["clono", "v",  "j"]
        tmp_df = tmp_df1.append(tmp_df2).append(tmp_df3)
        tmp_df.loc[:, "cell_count"] = np.repeat(int(1), len(tmp_df))
        tmp_df = tmp_df.iloc[:, np.r_[0, 3, 1, 2]]
        clono_type_list = [x for x in tmp_df.clono.unique().
                           tolist() if str(x) != 'nan']
        for j in range(0, len(clono_type_list)):
            temp_df = tmp_df.loc[tmp_df.clono == 
                clono_type_list[j], :].drop(['clono'], axis=1)
            if len(temp_df) > 0:
                temp_df.loc[:,'vj'] = temp_df['v'].\
                    astype(str) + "&" + temp_df['j'].astype(str)
                temp_agg = temp_df.groupby('vj').\
                    agg({'cell_count': sum}).reset_index()
                temp_agg = temp_agg.iloc[:, np.r_[1, 0]]
                temp_vj = pd.concat([temp_agg['cell_count'], 
                    temp_agg['vj'].str.split('&', expand=True)], axis=1)
                temp_vj.columns = ["cell_count", "v", "j"]
                temp_vj.to_csv(os.path.join(file_dir, cell_type[0] + 
                     "_" + clono_type_list[j] + "_vj_join.txt"),
                     header=True, index=None, sep='\t')

    write_log("info",
              "Function do_tertiary in directory " + file_dir +
              " Succesfully completed", module_logname)

    return True
