#!/usr/bin/env python3
"""Module: shared_utility"""

import logging
import sys
import getopt
import os.path
import csv
import subprocess


def write_log(logtype, message, logname, also_stdout = False):
    """Function: Writing a message to a logger

    logtype can take two values: info and error
    """
    # Connect to the logger
    logger = logging.getLogger(logname)
    
    # Write log
    if logtype == "info":
        logger.info(message)
        if also_stdout == True:
            print("INFO: ", message)
    elif logtype == "error":
        logger.error(message)
        if also_stdout == True:
            print("ERROR: ", message)
    else:
        exit_failure("Unknown logtype: " + logtype, logname)


def exit_success(message, logname):
    """Function: exit success

    Write message to logger
    Exit 0
    """
    write_log("info", message, logname)
    sys.exit(0)


def exit_failure(message, logname, exit_code = 1):
    """Function: exit success

    Write message to logger
    Exit 1, unless exit code provided
    """
    write_log("error", message, logname)
    sys.exit(int(exit_code))



def call_logger(mode, also_print_on_stdout, message, logname):
    """Function: Generic call to a specified logger.

    mode: Defines action to be taken and can take one of following values: 
        0:  successfull termination of program (last command of pipeline)
        2:  command line syntax error (typically user input error) and exit
        20: same as 2 but don't exit
        1:  all other errors and exit
        10: same as 1 but don't exit
        -1: info (including intermediate successes of functions/modules)
        all other values will result in an unrecognized mode exit.
    also_print_on_stdout: a flag which on True also prints error to stdout.
    logname: also recommended to follow python conventions.
    """

    # Connect to the parent logger
    logger = logging.getLogger(logname)

    # Process log call
    if mode == -1:
        logger.info(message)
        if also_print_on_stdout == True:
            print("INFO: ", message)
    elif mode == 0:
        logger.info(message)
        if also_print_on_stdout == True:
            print("INFO: ", message)
        sys.exit(0)
    elif mode == 2 or mode ==20:
        logger.error(message)
        if also_print_on_stdout == True:
            print("ERROR: ", message)
        if mode == 2:
            sys.exit(2)
    elif mode == 1 or mode == 10:
        logger.error(message)
        if also_print_on_stdout == True:
            print("ERROR: ", message)
        if mode == 1:
            sys.exit(1)
    else:
        logger.error("Unrecognized mode in call_logger: %s", mode)
        if also_print_on_stdout == True:
            print("ERROR: ", message)
        sys.exit(1)


def do_load_csv_byrow(filename, logname):
    """Function: General loader of a csv file.

    Data is assumed to stored row-wise.
    The first column is considered to be a keyword and is stored as the keyword
    in a dictionary. The remaining columns are stored in the form of a list
    and as the value of the keyword.
    The dictionary is returned at the end.
    """

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", 
            "Function do_load_csv_byrow: Started", module_logname)

    # Read config file, validate keywords and values
    metadata = {}
    try:
        with open(filename, newline='') as csvfile:
            configreader = csv.reader(csvfile, skipinitialspace = True)
            for row in configreader:
                if (len(row) > 0) and (not row[0].startswith('#')):
                    metadata[row[0]] = row[1:]
    except EnvironmentError as err:
        write_log("error", "Unable to open file: " + csvfile + "\n" + str(err),
                module_logname)

    
    #print(metadata)
    write_log("info", "Function do_load_csv_byrow: Succesfully completed",
            module_logname)
    return metadata


def create_dir(dirname, logname):
    """Function: Create a directory given the location.

    Covers most failure modes using the latest in-built function.
    """

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function create_dir: Started", module_logname)

    try:
        os.makedirs(dirname)
    except OSError as err:
        write_log("error", "Unable to create directory: " + dirname, module_logname)
        return False


    write_log("info", "Function create_dir: Succesfully completed",
            module_logname)
    return True


def run_system_cmd(cmd, outfile, errfile, err_message, logname, append_mode = False):
    """Function: Run a system command.

    Stores stdout in outfile and stderr in errfile
    err_message is a message to be logged if the command fails
    """

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function run_system_cmd: Started", module_logname)

    if append_mode == True:
        with open(outfile, 'a') as out, open(errfile, 'a') as err:
            try:
                run_cmd = subprocess.run(cmd.split(), stdout=out, stderr=err, check=True)
            except subprocess.CalledProcessError as err:
                exit_failure(err_message + str(err), module_logname)

    else:
        with open(outfile, 'w') as out, open(errfile, 'w') as err:
            try:
                run_cmd = subprocess.run(cmd.split(), stdout=out, stderr=err, check=True)
            except subprocess.CalledProcessError as err:
                exit_failure(err_message + str(err), module_logname)

    write_log("info", "Function run_system_cmd: Succesfully completed",
            module_logname)


def check_config_value(key, val, rule, logname):
    """Function: Validates the value of a config using specified rules
    """
    # Connect to the parent logger
    module_logname = logname + "." + __name__

    if rule == "file":
        if not os.path.isfile(val):
            write_log("error", 
                "File does not exist: " + val, module_logname, True)
            return False
    elif rule == "dir":
        if not os.path.isdir(val):
            write_log("error",
                    "Path does not exist: " + val, module_logname, True)
            return False
    elif rule == "outdir":
        if os.path.isdir(val):
            write_log("error",
                    "Analysis dir already exists: " + val, module_logname, True)
            return False
        if(not create_dir(val, logname)):
            return False
    else:
        write_log("info", 
                "Config rule " + rule + " for keyword " + key + " ignored",
                module_logname)

    return True


def do_validate_config(configfile, valid_configs, opt_configs, logname):
    """Function: Validate the contents of a config file.

    Uses the keywords and rules as specified in valid_configs.
    Returns a dict with the user_configs."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function do_validate_config: Started", module_logname)

    # Read config file, validate keywords and values
    user_configs = {}
    with open(configfile, newline='') as csvfile:
        configreader = csv.reader(csvfile, skipinitialspace = True)
        for row in configreader:
            if (len(row) > 0) and (not row[0].startswith('#')):
                if row[0] not in valid_configs:
                    exit_failure("Invalid config keyword: " + row[0], 
                            module_logname, 2)
                elif(not check_config_value(row[0], row[1], 
                    valid_configs[row[0]], logname)):
                    exit_failure("Invalid value for config keyword: " + row[0], 
                            module_logname, 2)
                else:
                    if(row[0] in user_configs):
                        exit_failure( "Duplicate entries for config keyword: " + row[0],
                                module_logname, 2)
                    else:
                        user_configs[row[0]] = row[1]

    # Check required/ optional keywords 
    for key, val in valid_configs.items():
        if(key in user_configs):
            pass
        else:
            if (key in opt_configs):
                user_configs[key] = opt_configs[key]
            else:
                exit_failure("Required config " + key + " missing", 
                        module_logname, 2)
    
    # Save a copy of final user configs for reference
    if('analysis_dir' in user_configs):
        with open(os.path.join(user_configs['analysis_dir'],'config.bak'),'w') as out:
            for key, val in user_configs.items():
                out.write(key + ',' + val + '\n')


    #print(user_configs)
    write_log("info", 
            "Function do_validate_config: Succesfully completed", module_logname)
    return user_configs



def do_validate_arguments(sys_args, valid_options, valid_long_options, usage, logname):
    """Function: Validate command line arguments.

    Parses command line arguments. Validates config file argument.
    Returns config file path."""

    # Connect to the parent logger
    module_logname = logname + "." + __name__
    write_log("info", "Function do_validate_arguments: Started",
            module_logname)

    # Extract arguments and options
    try:
        opts, args = getopt.getopt(sys_args, valid_options, [valid_long_options])
    except getopt.GetoptError as err:
        write_log("error", usage, module_logname, True)
        exit_failure(
                "Unable to obtain sys arguments\n" + str(err), module_logname, 2)
    if len(sys_args) == 0:
        write_log("error", usage, module_logname, True)
        exit_failure("No arguments entered", module_logname, 2)
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            write_log("error", usage, module_logname, True)
            exit_failure("Check arguments", module_logname, 2)
        elif opt in ("-c", "--config"):
            config = arg
        else:
            write_log("error", usage, module_logname, True)
            exit_failure("Check arguments", module_logname, 2)

    if not os.path.isfile(config):
        write_log("error", usage, module_logname, True)
        exit_failure(
                "File does not exist: " + config, module_logname, 2)


    write_log("info", "Function do_validate_arguments: Succesfully completed", 
            module_logname)
    return config



