#!/usr/bin/env python3

# main script for ICELL8 single cell TCR a/b profiling analysis

import sys
import os
import logging
import re
from collections import OrderedDict
import platform

# Import from custom library
from lib.shared_utility import *
from lib.sc_tcr_utility import *

###############################################################################
# Main Script
###############################################################################

if __name__ == "__main__":
    """The main entry point of script
    """

    ###########################################################################
    # "CONSTANT" declarations
    ###########################################################################
    USAGE = "Usage: sc_tcr_analysis.py -c <configfile>"
    
    LOG_NAME = "sc_tcr_analysis"
    LOG_FILENAME = "sc_tcr_analysis.log"
    LOG_FORMAT = '%(levelname)s - %(asctime)s - %(name)s - %(message)s'
    LOG_FORMAT_ARG_DATE = "%Y-%m-%d %H:%M:%S"
    LOG_KEYWORD_INFO = "info"
    LOG_KEYWORD_ERROR = "error"

    VALID_OPTS = "hc:"
    VALID_OPTS_LONG = '"help","config="'

    KEYWORD_READ1_FASTQ = "read1_fastq"
    KEYWORD_READ2_FASTQ = "read2_fastq"
    KEYWORD_SELECTED_BARCODE_LIST = "selected_barcode_list"
    KEYWORD_PRINTED_BARCODE_LIST = "printed_barcode_list"
    KEYWORD_SELECTED_ONLY = "selected_only"
    KEYWORD_IMMU_LOCI = "immu_loci"
    KEYWORD_SPECIES = "species"
    KEYWORD_DEMUX_TOOL_LOC = "demux_tool_loc"
    KEYWORD_MIXCR_LOC = "mixcr_loc"
    KEYWORD_ANALYSIS_DIR = "analysis_dir"
    KEYWORD_REPORT_NAME = "experiment_name"
    KEYWORD_NEG_CTRL_NAME = "neg_ctrl_name"

    VALID_CONFIGS = {
        KEYWORD_READ1_FASTQ: "file",
        KEYWORD_READ2_FASTQ: "file",
        KEYWORD_SELECTED_BARCODE_LIST: "file",
        KEYWORD_PRINTED_BARCODE_LIST: "file",
        KEYWORD_SELECTED_ONLY: "NA",
        KEYWORD_IMMU_LOCI: "NA",
        KEYWORD_SPECIES: "NA",
        KEYWORD_DEMUX_TOOL_LOC: "file",
        KEYWORD_MIXCR_LOC: "file",
        KEYWORD_ANALYSIS_DIR: "outdir",
        KEYWORD_REPORT_NAME: "NA",
        KEYWORD_NEG_CTRL_NAME: "NA"}

    OPTIONAL_CONFIGS = {
        KEYWORD_SELECTED_ONLY: "TRUE",
        KEYWORD_REPORT_NAME: "sc_tcr",
        KEYWORD_NEG_CTRL_NAME: "NegCtrl",
        KEYWORD_IMMU_LOCI: "TRA,TRB",
        KEYWORD_SPECIES: "hsa"}
    
    # name subdirectory
    SPLIT_FASTQ_SUBDIR = "split_fastqs" # contains barcode-level fastqs
    ALIGN_SUBDIR = "align"              # contains *_align_report.txt & *.vdjca
    ASSEMBLE_SUBDIR = "assemble_clones" # contains *_clones_report.txt & *.clns
    ALL_CLONES_SUBDIR = "export_clones" # contains *_clones_all.txt
    VDJVIZ_SUBDIR = "vdjviz_input"

    ###########################################################################
    # Setup Logging
    ###########################################################################
    # Initiate logging and set level
    logger = logging.getLogger(LOG_NAME)
    logger.setLevel(logging.INFO)
    # Create file handle for logging
    fh = logging.FileHandler(LOG_FILENAME, mode="w")
    # Set format for logging
    formatter = logging.Formatter(LOG_FORMAT, datefmt=LOG_FORMAT_ARG_DATE)
    fh.setFormatter(formatter)
    # Add handler to logger object
    logger.addHandler(fh)

    write_log(LOG_KEYWORD_INFO, "Main Program: Started\n=======\n", LOG_NAME)
    print("Main Program: Started ======= ", flush = True) 

    #
    # Validate input command line arguments
    #
    configfile = do_validate_arguments(sys.argv[1:],
                                       VALID_OPTS, VALID_OPTS_LONG,
                                       USAGE, LOG_NAME)

    #
    # Validate config file contents
    # Format: keyword,value
    #
    user_configs = do_validate_config(configfile,
                                      VALID_CONFIGS, 
                                      OPTIONAL_CONFIGS, 
                                      LOG_NAME)
 
    if user_configs[KEYWORD_READ1_FASTQ] == user_configs[KEYWORD_READ2_FASTQ]:
        print("Please check: FASTQs for read1 and read2 should be different.")
        print("Your read1 & read2 FASTQs are:")
        print(user_configs[KEYWORD_READ1_FASTQ])
        print(user_configs[KEYWORD_READ2_FASTQ])
        print("Pipeline run is terminated.")
        write_log(LOG_KEYWORD_INFO,
            "Pipeline is terminated due to identical FASTQs for read1 & read2",
            LOG_NAME)
        sys.exit()

    #
    # load selected and/or printed barcode list, classify barcode source
    #
     
    # load printed barcode list: a python list with barcode string as elements 
    #   if selected_only is on, this step is skipped
    if user_configs[KEYWORD_SELECTED_ONLY] == "TRUE":
         pass
    else:
        write_log(LOG_KEYWORD_INFO,"Printed barcode list loading started",
                  LOG_NAME)
        printed_selected_barcode_list = []
        printed_selected_barcode_list = load_printed_barcode(
                                  user_configs[KEYWORD_PRINTED_BARCODE_LIST],
                                  LOG_NAME)
        write_log(LOG_KEYWORD_INFO, "Completed printed barcode list loading",
                  LOG_NAME)
    
    # load selected barcode list: a python dictionary
    #   with sample_id as key & with barcode as value 
    write_log(LOG_KEYWORD_INFO, "Selected barcode list loading started", 
              LOG_NAME)
    selected_barcode_dict = load_selected_barcode(
                                user_configs[KEYWORD_SELECTED_BARCODE_LIST],
                                LOG_NAME)
    write_log(LOG_KEYWORD_INFO, "Completed selected barcode list loading",
              LOG_NAME)

    # classify notselected barcode when selected_only is turned off
    if user_configs[KEYWORD_SELECTED_ONLY]=="TRUE":
        pass
    else:
        write_log(LOG_KEYWORD_INFO, "barcode classification started", 
                 LOG_NAME)

        full_barcode_dict = do_barcode_classification(
                                printed_selected_barcode_list,
                                selected_barcode_dict,
                                user_configs[KEYWORD_SELECTED_BARCODE_LIST],
                                LOG_NAME)                                

        write_log(LOG_KEYWORD_INFO, 
                  "Completed barcode classification", 
                  LOG_NAME)
   
    # Create sub-directory to store splited fastqs
    if (not create_dir(os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                           SPLIT_FASTQ_SUBDIR), LOG_NAME)):
        exit_failure("Unable to create split_fastq sub-dir", LOG_NAME)

    # Launch ICELL8TCR_Demux.jar to generate barcode-level fastqs  
    print("Start demultiplexing.....", flush = True) 
    if user_configs[KEYWORD_SELECTED_ONLY]=="TRUE":
        if (not split_fastq(user_configs[KEYWORD_DEMUX_TOOL_LOC],
                            user_configs[KEYWORD_READ2_FASTQ],
                            user_configs[KEYWORD_READ1_FASTQ],
                            os.path.splitext(user_configs[
                                KEYWORD_SELECTED_BARCODE_LIST])[0] + 
                                ".csv",
                            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                                SPLIT_FASTQ_SUBDIR),
                            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                                SPLIT_FASTQ_SUBDIR,
                                user_configs[KEYWORD_REPORT_NAME]) + 
                                "_demultiplexing.log",
                            LOG_NAME)):
                exit_failure("Unable to split fastqs", LOG_NAME)

    else:
        if (not split_fastq(user_configs[KEYWORD_DEMUX_TOOL_LOC],
                            user_configs[KEYWORD_READ2_FASTQ],
                            user_configs[KEYWORD_READ1_FASTQ],
                            os.path.splitext(user_configs[
                                KEYWORD_SELECTED_BARCODE_LIST])[0] + 
                                "_printed.csv",
                            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                                SPLIT_FASTQ_SUBDIR),
                            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                                SPLIT_FASTQ_SUBDIR,
                                user_configs[KEYWORD_REPORT_NAME]) +
                                "_demultiplexing.log",
                            LOG_NAME)):
                exit_failure("Unable to split fastqs", LOG_NAME)
    print("Completed demultiplexing", flush = True) 

    # if selected_only, then clean intermediate fastq  
    # else, append intermediate fastq to full_barcode_dict
    if user_configs[KEYWORD_SELECTED_ONLY] == "TRUE":
        if (not clean_intermediates(
            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                         SPLIT_FASTQ_SUBDIR), 
            LOG_NAME)):
            exit_failure("Failed to clean intermediate file",LOG_NAME)  
    else:
        full_barcode_dict["INVALID"] = "NA"

    #
    # launch MiXCR procedures/workflows
    # 
 
    # Create MiXCR align sub-directory
    if (not create_dir(os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                           ALIGN_SUBDIR), LOG_NAME)):
        exit_failure("Unable to create MiXCR align sub-dir: " + key,
                     LOG_NAME)

    # Create MiXCR assemble sub-directory
    if (not create_dir(os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                           ASSEMBLE_SUBDIR), LOG_NAME)):
        exit_failure("Unable to create MiXCR assemble sub-dir: " + key,
                     LOG_NAME)

    # Create MiXCR all clones sub-directory
    if (not create_dir(os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                           ALL_CLONES_SUBDIR), LOG_NAME)):
        exit_failure("Unable to create MiXCR all clones sub-dir: " + key,
                     LOG_NAME)

    # Create VDJviz imput sub-directory
    if (not create_dir(os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                           VDJVIZ_SUBDIR), LOG_NAME)):
        exit_failure("Unable to create VDJviz input sub-dir: " + key,
                     LOG_NAME)

    # based on selected_only switch, 
    # use selected barcode dict or full barcode dict as final dict
    if user_configs[KEYWORD_SELECTED_ONLY] == "TRUE":
        final_barcode_dict = selected_barcode_dict
    else:
        final_barcode_dict = full_barcode_dict

    # Iterate over barcode list to launch MiXCR workflow components
    write_log(LOG_KEYWORD_INFO,
              "Iteration over barcode list to launch MiXCR workflows",
              LOG_NAME)

    for key, val in final_barcode_dict.items():
        write_log(LOG_KEYWORD_INFO, "MiXCR starts for " + key + "-" + val,
                  LOG_NAME)
        print("Processing " + key + " ......", flush=True)

        # add key,value pair MiXCR process info to stdout
        if (not add_info_to_file(
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                    user_configs[KEYWORD_REPORT_NAME]) + "_mixcr.stdout",
                "============= MiXCR for ",
                key + "-" + val,
                " being processed ============= \n",
                LOG_NAME)):
            exit_failure("Unable to add sample " + key + " to stdout",
                         LOG_NAME)
        
        # check os, if windows convert dos format fastq files to unix format
        if platform.system() == "Windows":
            if (not dos2unix(
                    os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], SPLIT_FASTQ_SUBDIR,
                        key + "_" + os.path.basename(
                        user_configs[KEYWORD_READ1_FASTQ]).replace(".gz","")),
                LOG_NAME)):
                exit_failure("Unable to dos2unix fastq1 for: " + key, LOG_NAME)

            if (not dos2unix(
                    os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], SPLIT_FASTQ_SUBDIR,
                        key + "_" + os.path.basename(
                        user_configs[KEYWORD_READ2_FASTQ]).replace(".gz","")),
                LOG_NAME)):
                exit_failure("Unable to dos2unix fastq2 for: " + key, LOG_NAME)
        else:
            pass
            
        # MiXCR alignment
        # aligns raw sequencing reads to reference V, D, J and C genes of
        #  T- and B- cell receptors.

        if (not mixcr_align(
                user_configs[KEYWORD_MIXCR_LOC],
                user_configs[KEYWORD_IMMU_LOCI],
                user_configs[KEYWORD_SPECIES],
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    SPLIT_FASTQ_SUBDIR,
                    key + "_" + os.path.basename(
                    user_configs[KEYWORD_READ1_FASTQ]).replace(".gz","")),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    SPLIT_FASTQ_SUBDIR,
                    key + "_" + os.path.basename(
                    user_configs[KEYWORD_READ2_FASTQ]).replace(".gz","")),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    ALIGN_SUBDIR,
                    key + "-" + val + ".vdjca"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    ALIGN_SUBDIR,
                    key + "-" + val + "_align_report.txt"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    user_configs[KEYWORD_REPORT_NAME] + "_mixcr.stdout"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                    user_configs[KEYWORD_REPORT_NAME] + "_mixcr.stderr"),
                LOG_NAME)):
            exit_failure("Unable to do mixcr_align for: " + key, LOG_NAME)

        # MiXCR assemble clones
        # builds a set of clones using alignments obtained with align command
        #   in order to extract specific gene regions (e.g. CDR3).
        if (not mixcr_assemble(
                user_configs[KEYWORD_MIXCR_LOC],
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    ALIGN_SUBDIR,
                    key + "-" + val + ".vdjca"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    ASSEMBLE_SUBDIR,
                    key + "-" + val + ".clns"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    ASSEMBLE_SUBDIR,
                    key + "-" + val + "_clones_report.txt"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                    user_configs[KEYWORD_REPORT_NAME] + "_mixcr.stdout"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                    user_configs[KEYWORD_REPORT_NAME] + "_mixcr.stderr"),
                LOG_NAME)):
            exit_failure("Unable to do mixcr_assemble for: " + key, LOG_NAME)

        # MiXCR export clones
        # export result of alignment or clones from binary file
        #  (.vdjca or .clns) to a human-readable text file
        # note here: 0% Exporting clones would be redirect to stdout 
        #   instead of MiXCR stderr
        if (not mixcr_export_clones(
                user_configs[KEYWORD_MIXCR_LOC],
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    ASSEMBLE_SUBDIR,
                    key + "-" + val + ".clns"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                    ALL_CLONES_SUBDIR,
                    key + "-" + val + "_clones_all.txt"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                    user_configs[KEYWORD_REPORT_NAME] + "_mixcr.stderr"),
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                    user_configs[KEYWORD_REPORT_NAME] + "_mixcr.stdout"),
                LOG_NAME)):
            exit_failure("Unable to do mixcr_export_clones for: " + key,
                         LOG_NAME)
       
        # add MiXCR completion info for key,value pair to stdout
        if (not add_info_to_file(
                os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
                    user_configs[KEYWORD_REPORT_NAME] + "_mixcr.stdout"),
                "============= MiXCR for ",
                key + "-" + val,
                " is completed ============= \n\n",
                LOG_NAME)):
            exit_failure("Unable to add " + key + " to stdout", LOG_NAME)
        
        print("Completed " + key, flush=True)
        
        vdjviz_convert(
            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                ALL_CLONES_SUBDIR,
                key + "-" + val + "_clones_all.txt"),
            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR], 
                VDJVIZ_SUBDIR,
                key + "-" + val + "_vdjviz.txt"))

    write_log(LOG_KEYWORD_INFO, "MiXCR procedures are completed", LOG_NAME)
    
    #
    # start MiXCR results summarization & reads counting
    #
    
    print("Summarizing whole chip results ......", flush=True)
    # write header for *summary.csv
    write_summary_header(os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
        user_configs[KEYWORD_REPORT_NAME] + "_summary.csv"))
    
    # initiate global variables to count reads
    initiate_global_count()
    
    # Iterate over barcode list to summarize barcode-level results 
    write_log(LOG_KEYWORD_INFO,
              "Iteration over barcode list to summarize results",
              LOG_NAME)

    for key, val in final_barcode_dict.items():
        write_log(LOG_KEYWORD_INFO, "Summarize results for " +
                  key + "-" + val,
                  LOG_NAME)
       
        # summarize mixcr results
        summarize_mixcr_results_by_sample(user_configs[KEYWORD_ANALYSIS_DIR], 
            key + "-" + val, user_configs[KEYWORD_REPORT_NAME], ALIGN_SUBDIR, 
            ASSEMBLE_SUBDIR, ALL_CLONES_SUBDIR)
        
    write_log(LOG_KEYWORD_INFO,
              "Completed mixcr results iteration",
              LOG_NAME)

    # count reads in barcode-level fastqs
    if user_configs[KEYWORD_SELECTED_ONLY] == "TRUE":
        pass     
    else:
        count_read_in_fastq(user_configs[KEYWORD_ANALYSIS_DIR], 
                            SPLIT_FASTQ_SUBDIR,
                            final_barcode_dict,
                            user_configs[KEYWORD_READ1_FASTQ],
                            user_configs[KEYWORD_REPORT_NAME])

    # count barcode in intermediate fastq; skip if selected_only is turned off
    if user_configs[KEYWORD_SELECTED_ONLY] == "TRUE":
        pass       
    else:
        count_barcode_in_fastq(
            os.path.join(user_configs[KEYWORD_ANALYSIS_DIR],
            SPLIT_FASTQ_SUBDIR,           
            "INVALID_" +
            os.path.basename(user_configs[
                KEYWORD_READ1_FASTQ]).replace(".gz","")),
            user_configs[KEYWORD_ANALYSIS_DIR],
            user_configs[KEYWORD_REPORT_NAME])
    
    # 
    # record stats for data set and clear up files
    # 
    write_log(LOG_KEYWORD_INFO,
              "Record data set stats and clear up files ",
              LOG_NAME)

    # record data set stats in *summary.csv
    if user_configs[KEYWORD_SELECTED_ONLY] == "TRUE":
        write_all_samples_brief_stats(user_configs[KEYWORD_ANALYSIS_DIR],
                                     user_configs[KEYWORD_REPORT_NAME])
    else:
        write_all_samples_stats(user_configs[KEYWORD_ANALYSIS_DIR],
                                user_configs[KEYWORD_REPORT_NAME])    

    # tertiary analysis of summary.csv
    if (not do_tertiary(user_configs[KEYWORD_ANALYSIS_DIR],
                        user_configs[KEYWORD_REPORT_NAME],
                        user_configs[KEYWORD_NEG_CTRL_NAME],
                        LOG_NAME)):
        exit_failure("Unable to perform tertiary analysis",LOG_NAME)    

    print("Successfully completed =======", flush=True)

    # finish
    exit_success("Main Program: Succesfully completed\n=======\n", LOG_NAME)


